import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core.enums import Optimizer, Initializer, Activation
from neuralogic.core.error_function import MSE, SoftEntropy, CrossEntropy, ErrorFunction


class SettingsProxy:
    def __init__(
        self,
        *,
        optimizer: Optimizer,
        learning_rate: float,
        epochs: int,
        error_function: ErrorFunction,
        initializer: Initializer,
        initializer_const: float,
        initializer_uniform_scale: float,
        rule_neuron_activation: Activation,
        relation_neuron_activation: Activation,
        iso_value_compression: bool,
        chain_pruning: bool,
    ):
        if not is_initialized():
            initialize()

        self.settings_class = jpype.JClass("cz.cvut.fel.ida.setup.Settings")
        self.settings = self.settings_class()

        params = locals().copy()
        params.pop("self")

        for key, value in params.items():
            self.__setattr__(key, value)

        self.settings.debugExporting = False
        self.settings.exportBlocks = []

        self.settings.infer()

    @property
    def iso_value_compression(self) -> bool:
        return self.settings.isoValueCompression

    @iso_value_compression.setter
    def iso_value_compression(self, iso_value_compression: bool):
        self.settings.isoValueCompression = iso_value_compression

    @property
    def chain_pruning(self) -> bool:
        return self.settings.chainPruning

    @chain_pruning.setter
    def chain_pruning(self, chain_pruning: bool):
        self.settings.chainPruning = chain_pruning

    @property
    def seed(self) -> int:
        return self.settings.seed

    @seed.setter
    def seed(self, seed: int):
        self.settings.seed = seed

    @property
    def learning_rate(self) -> float:
        return self.settings.initLearningRate

    @learning_rate.setter
    def learning_rate(self, learning_rate: float):
        self.settings.initLearningRate = learning_rate

    @property
    def optimizer(self):
        return self.settings.getOptimizer()

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        if optimizer == Optimizer.SGD:
            java_optimizer = self.settings_class.OptimizerSet.SGD
        elif optimizer == Optimizer.ADAM:
            java_optimizer = self.settings_class.OptimizerSet.ADAM
        else:
            raise NotImplementedError
        self.settings.setOptimizer(java_optimizer)

    @property
    def initializer_const(self):
        return self.settings.constantInitValue

    @initializer_const.setter
    def initializer_const(self, value: float):
        self.settings.constantInitValue = value

    @property
    def initializer_uniform_scale(self):
        return self.settings.randomInitScale

    @initializer_uniform_scale.setter
    def initializer_uniform_scale(self, value: float):
        self.settings.randomInitScale = value

    @property
    def error_function(self):
        return self.settings.errorFunction

    @error_function.setter
    def error_function(self, error_function: ErrorFunction):
        if isinstance(error_function, MSE):
            self.settings.squishLastLayer = False
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.REGRESSION
            java_error_function = self.settings_class.ErrorFcn.SQUARED_DIFF
        elif isinstance(error_function, SoftEntropy):
            self.settings.squishLastLayer = True
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.CLASSIFICATION
            java_error_function = self.settings_class.ErrorFcn.SOFTENTROPY
        elif isinstance(error_function, CrossEntropy):
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.CLASSIFICATION

            if error_function.with_logits:
                self.settings.squishLastLayer = True
                java_error_function = self.settings_class.ErrorFcn.SOFTENTROPY
            else:
                self.settings.squishLastLayer = False
                java_error_function = self.settings_class.ErrorFcn.CROSSENTROPY
        else:
            raise NotImplementedError

        self.settings.errorFunction = java_error_function

    @property
    def epochs(self) -> int:
        return self.settings.maxCumEpochCount

    @epochs.setter
    def epochs(self, epochs: int):
        self.settings.maxCumEpochCount = epochs

    @property
    def initializer(self):
        initializer = self.settings.initializer

        if str(initializer) != "SIMPLE":
            return initializer
        return self.settings.initDistribution

    @initializer.setter
    def initializer(self, initializer: Initializer):
        if initializer == Initializer.HE:
            self.settings.initializer = self.settings_class.InitSet.HE
            return
        if initializer == Initializer.GLOROT:
            self.settings.initializer = self.settings_class.InitSet.GLOROT
            return

        if initializer == Initializer.NORMAL:
            init_dist = self.settings_class.InitDistribution.NORMAL
        elif initializer == Initializer.UNIFORM:
            init_dist = self.settings_class.InitDistribution.UNIFORM
        elif initializer == Initializer.CONSTANT:
            init_dist = self.settings_class.InitDistribution.CONSTANT
        elif initializer == Initializer.LONGTAIL:
            init_dist = self.settings_class.InitDistribution.LONGTAIL
        else:
            raise NotImplementedError
        self.settings.initDistribution = init_dist
        self.settings.initializer = self.settings_class.InitSet.SIMPLE

    @property
    def relation_neuron_activation(self) -> Activation:
        return self.settings.atomNeuronActivation

    @relation_neuron_activation.setter
    def relation_neuron_activation(self, value: Activation):
        self.settings.atomNeuronActivation = self.get_activation_function(value)

    @property
    def rule_neuron_activation(self) -> Activation:
        return self.settings.ruleNeuronActivation

    @rule_neuron_activation.setter
    def rule_neuron_activation(self, value: Activation):
        self.settings.ruleNeuronActivation = self.get_activation_function(value)

    @property
    def debug_exporting(self) -> bool:
        return self.settings.debugExporting

    @debug_exporting.setter
    def debug_exporting(self, debug_export: bool):
        self.settings.debugExporting = debug_export

    @property
    def default_fact_value(self) -> float:
        return self.settings.defaultFactValue

    @default_fact_value.setter
    def default_fact_value(self, value: float):
        self.settings.defaultFactValue = value

    def get_activation_function(self, activation: Activation):
        if activation == Activation.SIGMOID:
            return self.settings_class.ActivationFcn.SIGMOID
        if activation == Activation.TANH:
            return self.settings_class.ActivationFcn.TANH
        if activation == Activation.SIGNUM:
            return self.settings_class.ActivationFcn.SIGNUM
        if activation == Activation.RELU:
            return self.settings_class.ActivationFcn.RELU
        if activation == Activation.IDENTITY:
            return self.settings_class.ActivationFcn.IDENTITY
        if activation == Activation.LUKASIEWICZ:
            return self.settings_class.ActivationFcn.LUKASIEWICZ
        if activation == Activation.SOFTMAX:
            return self.settings_class.ActivationFcn.SOFTMAX
        if activation == Activation.SPARSEMAX:
            return self.settings_class.ActivationFcn.SPARSEMAX
        raise NotImplementedError

    def to_json(self) -> str:
        return self.settings.exportToJson()



import json
from typing import Dict, Sized, Union, List

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.nn.base import AbstractNeuraLogic
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.enums import Backend


class Loss:
    def __init__(self, loss, number_format):
        self.loss = loss
        self.number_format = number_format

    def backward(self):
        self.loss.backward()

    def value(self) -> float:
        return json.loads(str(self.loss.getError().toString(self.number_format)))

    def output(self) -> Union[List[float], float]:
        return json.loads(str(self.loss.getOutput().toString(self.number_format)))

    def target(self) -> Union[List[float], float]:
        return json.loads(str(self.loss.getTarget().toString(self.number_format)))


class NeuraLogic(AbstractNeuraLogic):
    @jpype.JImplements(jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.iteration.actions.PythonHookHandler"))
    class HookHandler:
        def __init__(self, module: "NeuraLogic"):
            self.module = module

        @jpype.JOverride
        def handleHook(self, hook, value):
            self.module.run_hook(hook, json.loads(value))

    def __init__(self, model, template, settings: SettingsProxy):
        super().__init__(Backend.JAVA, template, settings)

        if not is_initialized():
            initialize()

        python_strategy = jpype.JClass(
            "cz.cvut.fel.ida.neural.networks.computation.training.strategies.PythonTrainingStrategy"
        )

        self.do_train = True
        self.need_sync = False

        self.neural_model = model
        self.strategy = python_strategy(settings.settings, model)
        self.samples_len = 0

        self.hook_handler = NeuraLogic.HookHandler(self)
        self.reset_parameters()

    def reset_parameters(self):
        self.strategy.resetParameters()

    def train(self):
        self.do_train = True

    def test(self):
        self.do_train = False

    def set_training_samples(self, samples):
        self.samples_len = len(samples)
        self.strategy.setSamples(jpype.java.util.ArrayList(samples))

    def __call__(self, samples=None, train: bool = None, auto_backprop: bool = False, epochs: int = 1):
        self.hooks_set = len(self.hooks) != 0

        if self.hooks_set:
            self.strategy.setHooks(set(self.hooks.keys()), self.hook_handler)

        if train is not None:
            self.do_train = train

        if samples is None:
            results = self.strategy.learnSamples(epochs)
            deserialized_results = json.loads(str(results))

            return deserialized_results, self.samples_len

        if not isinstance(samples, Sized):
            if self.do_train:
                if auto_backprop:
                    result = self.strategy.learnSample(samples.java_sample)
                    return json.loads(str(result)), 1
            result = self.strategy.evaluateSample(samples.java_sample)
            return Loss(result, self.settings.settings_class.superDetailedNumberFormat)

        if self.do_train:
            results = self.strategy.learnSamples(
                jpype.java.util.ArrayList([sample.java_sample for sample in samples]), epochs
            )
            deserialized_results = json.loads(str(results))

            return deserialized_results, len(samples)

        results = self.strategy.evaluateSamples(jpype.java.util.ArrayList([sample.java_sample for sample in samples]))
        return json.loads(str(results))

    def state_dict(self) -> Dict:
        weights = self.neural_model.getAllWeights()
        weights_dict = {}

        for weight in weights:
            if weight.isLearnable:
                value = weight.value

                size = list(value.size())

                if len(size) == 0 or size[0] == 0:
                    weights_dict[weight.index] = value.value
                elif len(size) == 1 or size[0] == 1 or size[1] == 1:
                    weights_dict[weight.index] = list(value.values)
                else:
                    weights_dict[weight.index] = json.loads(value.toString())
        return {
            "weights": weights_dict,
        }

    def load_state_dict(self, state_dict: Dict):
        self.sync_template(state_dict, self.neural_model.getAllWeights())



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.enums import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class APPNPConv(Module):
    r"""
    Approximate Personalized Propagation of Neural Predictions layer from `"Predict then Propagate: Graph Neural Networks meet Personalized PageRank" <https://arxiv.org/abs/1810.05997>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{0}_i = \mathbf{x}_i

    .. math::
        \mathbf{x}^{k}_i = \alpha \cdot \mathbf{x}^0_i + (1 - \alpha) \cdot {agg}_{j \in \mathcal{N}(i)}(\mathbf{x}^{k - 1}_j)

    .. math::
        \mathbf{x}^{\prime}_i = act(\mathbf{x}^{K}_i)

    Where *act* is an activation function and *agg* aggregation function.


    The first part of the second equation that is ":math:`\alpha \cdot \mathbf{x}^0_i`" is expressed in the logic form as:

    .. code-block:: logtalk

        R.<output_name>__<k>(V.I) <= R.<feature_name>(V.I)[<alpha>].fixed()

    The second part of the second equation that is ":math:`(1 - \alpha) \cdot {agg}_{j \in \mathcal{N}(i)}(\mathbf{x}^{k - 1}_j)`" is expressed as:

    .. code-block:: logtalk

        R.<output_name>__<k>(V.I) <= (R.<output_name>__<k-1>(V.J)[1 - <alpha>].fixed(), R.<edge_name>(V.J, V.I))

    Examples
    --------

    The whole computation of this module (parametrized as :code:`APPNPConv("h1", "h0", "_edge", 3, 0.1, Activation.SIGMOID)`) is as follows:

    .. code:: logtalk

        metadata = Metadata(activation=Activation.IDENTITY, aggregation=Aggregation.SUM)

        (R.h1__1(V.I) <= R.h0(V.I)[0.1].fixed()) | metadata
        (R.h1__1(V.I) <= (R.h0(V.J)[0.9].fixed(), R._edge(V.J, V.I))) | metadata
        R.h1__1/1 [Activation.IDENTITY]

        (R.h1__2(V.I) <= <0.1> R.h0(V.I)) | metadata
        (R.h1__2(V.I) <= (<0.9> R.h1__1(V.J), R._edge(V.J, V.I))) | metadata
        R.h1__2/1 [Activation.IDENTITY]

        (R.h1(V.I) <= <0.1> R.h0(V.I)) | metadata
        (R.h1(V.I) <= (<0.9> R.h1__2(V.J), R._edge(V.J, V.I))) | metadata
        R.h1 / 1 [Activation.SIGMOID]


    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    k : int
        Number of iterations
    alpha : float
        Teleport probability
    activation : Activation
        Activation function of the output.
        Default: ``Activation.IDENTITY``
    aggregation : Aggregation
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``

    """

    def __init__(
        self,
        output_name: str,
        feature_name: str,
        edge_name: str,
        k: int,
        alpha: float,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.alpha = alpha
        self.k = k

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)
        metadata = Metadata(activation=Activation.IDENTITY, aggregation=self.aggregation)
        edge = R.get(self.edge_name)
        feature = R.get(self.feature_name)

        rules = []
        for k in range(1, self.k):
            k_head = R.get(f"{self.output_name}__{k}")(V.I)
            rules.append((k_head <= feature(V.I)[self.alpha].fixed()) | metadata)

            if k == 1:
                rules.append((k_head <= (feature(V.J)[1 - self.alpha].fixed(), edge(V.J, V.I))) | metadata)
            else:
                rules.append(
                    (k_head <= (R.get(f"{self.output_name}__{k - 1}")(V.J)[1 - self.alpha].fixed(), edge(V.J, V.I)))
                    | metadata
                )
            rules.append(R.get(f"{self.output_name}__{k}") / 1 | Metadata(activation=Activation.IDENTITY))

        if self.k == 1:
            output_rule = head <= (feature(V.J)[1 - self.alpha].fixed(), edge(V.J, V.I))
        else:
            output_rule = head <= (
                R.get(f"{self.output_name}__{self.k - 1}")(V.J)[1 - self.alpha].fixed(),
                edge(V.J, V.I),
            )

        return [
            *rules,
            (head <= feature(V.I)[self.alpha].fixed()) | metadata,
            output_rule | metadata,
            R.get(self.output_name) / 1 | Metadata(activation=self.activation),
        ]



from neuralogic.core.enums import Activation, Aggregation, ActivationAgg, ActivationAggregation
from neuralogic.core.constructs.metadata import Metadata
from typing import Iterable, Optional


class Rule:
    def __init__(self, head, body):
        from neuralogic.core import Relation

        self.head = head

        if not isinstance(body, Iterable):
            body = [body]

        self.body = list(body)
        self.metadata: Optional[Metadata] = None

        if self.is_ellipsis_templated():
            variable_set = {term for term in head.terms if term is not Ellipsis and str(term)[0].isupper()}

            for body_atom in self.body:
                if body_atom.predicate.special and body_atom.predicate.name == "alldiff":
                    continue

                for term in body_atom.terms:
                    if term is not Ellipsis and str(term)[0].isupper():
                        variable_set.add(term)

            for atom_index, body_atom in enumerate(self.body):
                if not body_atom.predicate.special or body_atom.predicate.name != "alldiff":
                    continue

                new_terms = []
                found_replacement = False

                for index, term in enumerate(body_atom.terms):
                    if term is Ellipsis:
                        if found_replacement:
                            raise NotImplementedError
                        found_replacement = True
                        new_terms.extend(variable_set)
                    else:
                        new_terms.append(term)
                if found_replacement:
                    self.body[atom_index] = Relation.special.alldiff(*new_terms)

    def is_ellipsis_templated(self) -> bool:
        for body_atom in self.body:
            if not body_atom.predicate.special or body_atom.predicate.name != "alldiff":
                continue
            for term in body_atom.terms:
                if term is Ellipsis:
                    return True
        return False

    def __str__(self):
        metadata = "" if self.metadata is None is None else f" {self.metadata}"
        return f"{self.head.to_str()} :- {', '.join(atom.to_str() for atom in self.body)}.{metadata}"

    def __and__(self, other) -> "Rule":
        if isinstance(other, Iterable):
            self.body.extend(list(other))
        else:
            self.body.append(other)
        return self

    def __or__(self, other) -> "Rule":
        if isinstance(other, Iterable):
            metadata = Metadata()

            for entry in other:
                if isinstance(entry, (Activation, ActivationAgg, ActivationAggregation)):
                    metadata.activation = entry
                elif isinstance(entry, Aggregation):
                    metadata.aggregation = entry
                else:
                    raise NotImplementedError
            other = metadata
        elif not isinstance(other, Metadata):
            raise NotImplementedError
        self.metadata = other

        return self



from typing import Optional

from neuralogic.core.enums import Backend
from neuralogic.core.settings import Settings


def get_neuralogic_layer(backend: Backend = Backend.JAVA):
    if backend == Backend.DYNET:
        from neuralogic.nn.dynet import NeuraLogic  # type: ignore

        return NeuraLogic
    if backend == Backend.JAVA:
        from neuralogic.nn.java import NeuraLogic  # type: ignore

        return NeuraLogic
    if backend == Backend.PYG:
        from neuralogic.nn.native.pyg import NeuraLogic

        return NeuraLogic
    if backend == Backend.TORCH:
        from neuralogic.nn.torch import NeuraLogic  # type: ignore

        return NeuraLogic
    raise NotImplementedError


def get_evaluator(
    template: "Template",
    backend: Backend = Backend.JAVA,
    settings: Optional[Settings] = None,
):
    if settings is None:
        settings = Settings()

    if backend == Backend.DYNET:
        from neuralogic.nn.evaluators.dynet import DynetEvaluator

        return DynetEvaluator(template, settings)
    if backend == Backend.JAVA:
        from neuralogic.nn.evaluators.java import JavaEvaluator

        return JavaEvaluator(template, settings)
    if backend == Backend.PYG:
        from neuralogic.nn.evaluators.pyg import PyGEvaluator

        return PyGEvaluator(template, settings)
    if backend == Backend.TORCH:
        from neuralogic.nn.evaluators.torch import TorchEvaluator

        return TorchEvaluator(template, settings)



from typing import List

from neuralogic.core import Relation, Template, Var, Metadata, Activation, Aggregation
from neuralogic.utils.templates.modules import AbstractModule


class GINConv(AbstractModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        activation: Activation = Activation.RELU,
        aggregation: Aggregation = Aggregation.SUM,
        name=None,
        has_edge_attrs=True,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            activation=activation,
            aggregation=aggregation,
            name=name,
            has_edge_attrs=has_edge_attrs,
        )

    def build(
        self, template: Template, layer_count: int, previous_names: List[str], feature_name: str, edge_name: str
    ) -> str:
        name = f"l{layer_count}_gin" if self.name is None else self.name
        embed_name = f"{name}_embed"
        previous_name = feature_name if len(previous_names) == 0 else previous_names[-1]

        head_atom = Relation.get(embed_name)(Var.X)

        layer = head_atom <= (Relation.get(previous_name)(Var.Y), Relation.get(edge_name)(Var.Y, Var.X))
        template.add_rule(layer | Metadata(aggregation=Aggregation.SUM, activation=Activation.IDENTITY))
        template.add_rule((head_atom <= Relation.get(previous_name)(Var.X)) | Metadata(activation=Activation.IDENTITY))
        template.add_rule(Relation.get(embed_name) / 1 | Metadata(activation=Activation.IDENTITY))

        gin_head = Relation.get(name)

        layer = gin_head(Var.X)[self.out_channels, self.in_channels] <= head_atom[self.in_channels, self.in_channels]
        template.add_rule(layer | Metadata(activation=self.activation))
        template.add_rule(gin_head / 1 | Metadata(activation=self.activation))

        return name



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.enums import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class GATv2Conv(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        share_weights: bool = False,
        activation: Activation = Activation.IDENTITY,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.share_weights = share_weights
        self.activation = activation

    def __call__(self):
        w1 = f"{self.output_name}__right"
        w2 = w1 if self.share_weights else f"{self.output_name}__left"

        attention_metadata = Metadata(activation=Activation.LEAKY_RELU)
        metadata = Metadata(activation="product-identity", aggregation=Aggregation.SUM)
        attention = R.get(f"{self.output_name}__attention")

        head = R.get(self.output_name)
        feature = R.get(self.feature_name)
        edge = R.get(self.edge_name)

        return [
            (
                attention(V.I, V.J)[self.out_channels, self.out_channels]
                <= (
                    feature(V.I)[w2 : self.out_channels, self.in_channels],
                    feature(V.J)[w1 : self.out_channels, self.in_channels],
                )
            )
            | attention_metadata,
            attention / 2 | Metadata(activation=Activation.SOFTMAX),
            (head(V.I) <= (attention(V.I, V.J), feature(V.J)[w1 : self.out_channels, self.in_channels], edge(V.J, V.I)))
            | metadata,
            head / 1 | Metadata(activation=self.activation),
        ]



from enum import Enum


from neuralogic.core.error_function import SoftEntropy, CrossEntropy, MSE


class Optimizer(str, Enum):
    ADAM = "ADAM"
    SGD = "SGD"


class Activation(str, Enum):
    SIGMOID = "SIGMOID"
    TANH = "TANH"
    SIGNUM = "SIGNUM"
    RELU = "RELU"
    LEAKY_RELU = "LEAKYRELU"
    IDENTITY = "IDENTITY"
    LUKASIEWICZ = "LUKASIEWICZ"
    SOFTMAX = "SOFTMAX"
    SPARSEMAX = "SPARSEMAX"

    def __add__(self, other: "ActivationAgg"):
        if not isinstance(other, ActivationAgg):
            raise NotImplementedError
        return ActivationAggregation(other, self)

    def __str__(self):
        return self.value.lower()


class ActivationAgg(str, Enum):
    MAX = "max"
    MIN = "min"

    def __add__(self, other: Activation):
        if not isinstance(other, Activation):
            raise NotImplementedError
        return ActivationAggregation(self, other)

    def __str__(self):
        return f"{self.value}-identity"


class Aggregation(str, Enum):
    SUM = "sum"
    MAX = "max"
    AVG = "avg"
    MIN = "min"


class Initializer(str, Enum):
    UNIFORM = "UNIFORM"
    NORMAL = "NORMAL"
    CONSTANT = "CONSTANT"
    LONGTAIL = "LONGTAIL"
    GLOROT = "GLOROT"
    HE = "HE"


class Backend(Enum):
    DYNET = "dynet"
    JAVA = "java"
    PYG = "pyg"
    TORCH = "torch"


class ActivationAggregation:
    def __init__(self, aggregation: ActivationAgg, activation: Activation):
        self.aggregation = aggregation
        self.activation = activation

    def __str__(self):
        return f"{self.aggregation.value}-{self.activation.value.lower()}"



import torch
import torch_geometric

from neuralogic.core.enums import Activation, Aggregation
from neuralogic.utils.templates import TemplateList, GINConv, SAGEConv, GCNConv, GlobalPooling

native_activations = {
    str(Activation.RELU): torch.relu,
    str(Activation.TANH): torch.tanh,
    str(Activation.SIGMOID): torch.sigmoid,
    str(Activation.IDENTITY): lambda x: x,
    str(Activation.SOFTMAX): torch.softmax,
}

native_aggregations = {
    str(Aggregation.AVG): torch_geometric.nn.global_mean_pool,
    str(Aggregation.SUM): torch_geometric.nn.global_add_pool,
    str(Aggregation.MAX): torch_geometric.nn.global_max_pool,
}


class NeuraLogic(torch.nn.Module):
    def __init__(self, module_list: TemplateList):
        super().__init__()

        self.module_list = torch.nn.ModuleList()
        self.evaluations = []

        for i, module in enumerate(module_list.modules):
            activation = str(module.activation)
            if activation not in native_activations:
                raise Exception
            activation_fun = native_activations[activation]

            if isinstance(module, GCNConv):
                self.module_list.append(
                    torch_geometric.nn.GCNConv(
                        module.in_channels, module.out_channels, normalize=False, cached=False, bias=False
                    )
                )
            elif isinstance(module, SAGEConv):
                self.module_list.append(
                    torch_geometric.nn.SAGEConv(module.in_channels, module.out_channels, normalize=False, bias=False)
                )
            elif isinstance(module, GINConv):
                mlp = torch.nn.Sequential(
                    torch.nn.Linear(module.in_channels, module.out_channels),
                    torch.nn.ReLU(),
                    torch.nn.Linear(module.out_channels, module.out_channels),
                )

                self.module_list.append(torch_geometric.nn.GINConv(mlp))
            elif isinstance(module, GlobalPooling):
                pooling_layers = []

                for _ in module.jumping_knowledge:
                    layer = torch.nn.Linear(module.in_channels, module.out_channels, bias=False)
                    pooling_layers.append(layer)
                    self.module_list.append(layer)

                def _pooling(x, edge_index, batch, xs, i=i, module=module):
                    pooling_xs = [
                        self.module_list[i + j](native_aggregations[str(module.aggregation)](xs[layer], batch))
                        for j, layer in enumerate(module.jumping_knowledge)
                    ]

                    pooling_x = torch.stack(pooling_xs, dim=0)
                    return native_activations[str(module.activation)](torch.sum(pooling_x, dim=0))

                self.evaluations.append(_pooling)
                continue
            else:
                raise Exception

            def evaluate(x, edge_index, batch, xs, i=i, activation_fun=activation_fun):
                return activation_fun(self.module_list[i](x, edge_index))

            self.evaluations.append(evaluate)

    def forward(self, x, edge_index, batch=None):
        xs = []

        for evaluation in self.evaluations:
            x = evaluation(x, edge_index, batch, xs)
            xs.append(x)
        return x



from neuralogic.nn.module.module import Module
from neuralogic.nn.module.gcn import GCNConv
from neuralogic.nn.module.gin import GINConv
from neuralogic.nn.module.gsage import SAGEConv
from neuralogic.nn.module.rgcn import RGCNConv
from neuralogic.nn.module.tag import TAGConv
from neuralogic.nn.module.gatv2 import GATv2Conv
from neuralogic.nn.module.sg import SGConv
from neuralogic.nn.module.appnp import APPNPConv



import os
from typing import Optional, List

import jpype

from neuralogic.logging import _init_logging


_is_initialized = False

jvm_params = {
    "classpath": os.path.join(os.path.abspath(os.path.dirname(__file__)), "jar", "NeuraLogic.jar"),
}

jvm_options = ["-Xms1g", "-Xmx64g"]


def set_jvm_options(options: List[str]) -> None:
    """
    Set the jvm options - by default ["-Xms1g", "-Xmx64g"],
    """
    global jvm_options
    jvm_options = options


def set_jvm_path(path: Optional[str]) -> None:
    global jvm_params

    if path is None:
        jvm_params.pop("jvmpath", None)
    else:
        jvm_params["jvmpath"] = path


def is_initialized() -> bool:
    return _is_initialized


def initialize(
    debug_mode: bool = False, debug_port: int = 12999, is_debug_server: bool = True, debug_suspend: bool = False
):
    global _is_initialized

    if _is_initialized:
        raise Exception("NeuraLogic already initialized")

    _is_initialized = True
    if debug_mode:
        port = int(debug_port)
        server = "y" if is_debug_server else "n"
        suspend = "y" if debug_suspend else "n"

        debug_params = [
            "-Xint",
            "-Xdebug",
            "-Xnoagent",
            f"-Xrunjdwp:transport=dt_socket,server={server},address={port},suspend={suspend}",
        ]

        jpype.startJVM(*jvm_options, *debug_params, **jvm_params)
    else:
        jpype.startJVM(*jvm_options, **jvm_params)
    _init_logging()



from neuralogic.core.builder.builder import Builder
from neuralogic.core.builder.dataset_builder import DatasetBuilder
from neuralogic.core.builder.components import Neuron, Weight, Sample, BuiltDataset




from typing import Optional, Dict, Union

import torch
import torch.nn.functional

from neuralogic.core.error_function import ErrorFunctionNames
from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset, Dataset
from neuralogic.core.settings import Settings
from neuralogic.core.enums import Backend, Optimizer


class TorchEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: torch.optim.SGD(param, lr=rate),
        Optimizer.ADAM: lambda param, rate: torch.optim.Adam(param, lr=rate),
    }

    error_functions = {
        str(ErrorFunctionNames.MSE): torch.nn.MSELoss(),
        str(ErrorFunctionNames.SOFTENTROPY): torch.nn.CrossEntropyLoss(),
        # ErrorFunction.ABS_DIFF: lambda out, target: dy.abs(out - target),
        # ErrorFunction.CROSSENTROPY: lambda out, target: pass
    }

    def __init__(
        self,
        template: Template,
        settings: Settings,
    ):
        super().__init__(Backend.TORCH, template, settings)

    def train(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = str(self.settings.error_function)
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in TorchEvaluator.trainers:
            raise NotImplementedError
        if error_function not in TorchEvaluator.error_functions:
            raise NotImplementedError

        trainer = TorchEvaluator.trainers[optimizer](self.neuralogic_model.model, self.settings.learning_rate)
        error_function = TorchEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0
                for sample in dataset.samples:
                    trainer.zero_grad(set_to_none=True)

                    if isinstance(sample.target, (int, float)):
                        label = torch.tensor([sample.target], dtype=torch.float64, requires_grad=False)
                    else:
                        label = torch.tensor(sample.target, dtype=torch.float64, requires_grad=False)

                    graph_output = self.neuralogic_model(sample)
                    loss = error_function(graph_output, label)

                    try:
                        loss.backward()
                    except RuntimeError:
                        pass
                    trainer.step()

                    total_loss += loss.item()
                    seen_instances += 1
                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            with torch.no_grad():
                for sample in dataset.samples:
                    graph_output = self.neuralogic_model(sample)

                    if graph_output.size() == (1,):
                        yield sample.target, graph_output[0].item()
                    elif not graph_output.size():
                        yield sample.target, graph_output.item()
                    else:
                        yield sample.target, graph_output

        if generator:
            return _test()
        return list(_test())

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from typing import Union

from neuralogic.core.enums import Activation, Aggregation, ActivationAgg, ActivationAggregation


class Metadata:
    def __init__(
        self,
        offset=None,
        learnable: bool = None,
        activation: Union[str, Activation, ActivationAgg, ActivationAggregation] = None,
        aggregation: Aggregation = None,
        duplicit_grounding: bool = False,
    ):
        self.offset = offset
        self.learnable = learnable
        self.activation = activation
        self.aggregation = aggregation
        self.duplicit_grounding = duplicit_grounding

    def __str__(self):
        metadata_list = []
        if self.offset is not None:
            metadata_list.append(f"offset={self.offset}")
        if self.learnable is not None:
            metadata_list.append(f"learnable={str(self.learnable).lower()}")
        if self.activation is not None:
            metadata_list.append(f"activation={str(self.activation)}")
        if self.aggregation is not None:
            metadata_list.append(f"aggregation={self.aggregation.value}")
        return f"[{', '.join(metadata_list)}]"



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.enums import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class GINConv(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)[self.out_channels, self.in_channels]
        embed = R.get(f"embed__{self.output_name}")

        metadata = Metadata(activation=Activation.IDENTITY, aggregation=self.aggregation)

        return [
            (head <= (R.get(self.feature_name)(V.J), R.get(self.edge_name)(V.J, V.I))) | metadata,
            (embed(V.I) <= R.get(self.feature_name)(V.I)) | metadata,
            (head <= embed(V.I)[self.in_channels, self.in_channels]) | Metadata(activation=self.activation),
            embed / 1 | Metadata(activation=Activation.IDENTITY),
            R.get(self.output_name) / 1 | Metadata(activation=self.activation),
        ]



from typing import Sequence

from neuralogic.core.enums import Activation, Aggregation, ActivationAggregation, ActivationAgg
from neuralogic.core.constructs.metadata import Metadata


class Predicate:
    """WeightedPredicate"""

    def __init__(self, name, arity, hidden=False, special=False):
        if name.startswith("_"):
            name = name[1:]
            hidden = True

        self.name = name
        self.arity = arity
        self.hidden = hidden
        self.special = special

    def set_arity(self, arity):
        if self.arity == arity:
            return self

    def to_str(self):
        special = "@" if self.special else ""
        hidden = "*" if self.hidden else ""
        return f"{hidden}{special}{self.name}"

    def __str__(self):
        special = "@" if self.special else ""
        hidden = "*" if self.hidden else ""
        return f"{hidden}{special}{self.name}/{self.arity}"

    def __or__(self, other) -> "PredicateMetadata":
        if isinstance(other, Sequence):
            metadata = Metadata()

            for entry in other:
                if isinstance(entry, (Activation, ActivationAgg, ActivationAggregation)):
                    metadata.activation = entry
                elif isinstance(entry, Aggregation):
                    metadata.aggregation = entry
                else:
                    raise NotImplementedError
            other = metadata
        elif not isinstance(other, Metadata):
            raise NotImplementedError
        return PredicateMetadata(self, other)


class PredicateMetadata:
    def __init__(self, predicate: Predicate, metadata: Metadata):
        if metadata.aggregation is not None or metadata.learnable is not None:
            raise NotImplementedError

        self.predicate = predicate
        self.metadata = metadata

    def __str__(self):
        return f"{self.predicate} {self.metadata}"



from typing import Optional, Dict, Union

import dynet as dy
from neuralogic.core.error_function import ErrorFunctionNames

from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset, Dataset
from neuralogic.core.settings import Settings
from neuralogic.core.enums import Backend, Optimizer


class DynetEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: dy.SimpleSGDTrainer(param, learning_rate=rate),
        Optimizer.ADAM: lambda param, rate: dy.AdamTrainer(param, alpha=rate),
    }

    error_functions = {
        str(ErrorFunctionNames.MSE): lambda out, target: dy.squared_distance(out, target),
        # ErrorFunction.ABS_DIFF: lambda out, target: dy.abs(out - target),
        # ErrorFunction.CROSSENTROPY: lambda out, target: pass
    }

    def __init__(
        self,
        template: Template,
        settings: Settings,
    ):
        super().__init__(Backend.DYNET, template, settings)

    def train(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = str(self.settings.error_function)
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in DynetEvaluator.trainers:
            raise NotImplementedError
        if error_function not in DynetEvaluator.error_functions:
            raise NotImplementedError

        trainer = DynetEvaluator.trainers[optimizer](self.neuralogic_model.model, self.settings.learning_rate)
        error_function = DynetEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0

                dy.renew_cg(immediate_compute=False, check_validity=False)

                for sample in dataset.samples:
                    if isinstance(sample.target, list):
                        label = dy.inputTensor(sample.target)
                    else:
                        label = dy.scalarInput(sample.target)
                    graph_output = self.neuralogic_model(sample)

                    loss = error_function(graph_output, label)

                    total_loss += loss.value()
                    loss.backward()
                    trainer.update()
                    seen_instances += 1

                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for sample in dataset.samples:
                dy.renew_cg(immediate_compute=False, check_validity=False)

                graph_output = self.neuralogic_model(sample)
                if isinstance(sample.target, list):
                    label = dy.inputTensor(sample.target)
                else:
                    label = dy.scalarInput(sample.target)

                results = (label.value(), graph_output.value())
                yield results

        if generator:
            return _test()
        return list(_test())

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from typing import Dict, Optional, Union

from neuralogic.core.settings import Settings
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core import Template, Backend, BuiltDataset, SettingsProxy, Dataset

from neuralogic.utils.visualize import draw_model


class AbstractNeuraLogic:
    def __init__(self, backend: Backend, dataset_builder: DatasetBuilder, settings: SettingsProxy):
        self.need_sync = True

        self.template = dataset_builder.parsed_template
        self.dataset_builder = dataset_builder

        self.backend = backend
        self.settings = settings

        self.hooks_set = False
        self.hooks = {}

    def __call__(self, sample):
        raise NotImplementedError

    def build_dataset(self, dataset: Union[Dataset, BuiltDataset], file_mode: bool = False):
        return self.dataset_builder.build_dataset(dataset, self.backend, self.settings, file_mode)

    def set_hooks(self, hooks):
        self.hooks_set = len(hooks) != 0
        self.hooks = hooks

    def run_hook(self, hook: str, value):
        for callback in self.hooks[hook]:
            callback(value)

    def sync_template(self, state_dict: Optional[Dict] = None, weights=None):
        state_dict = self.state_dict() if state_dict is None else state_dict
        weights = self.template.getAllWeights() if weights is None else weights
        weight_dict = state_dict["weights"]

        for weight in weights:
            if not weight.isLearnable:
                continue
            weight_value = weight.value

            index = weight.index
            value = weight_dict[index]

            if isinstance(value, (float, int)):
                weight_value.set(0, float(value))
                continue

            if isinstance(value[0], (float, int)):
                for i, val in enumerate(value):
                    weight_value.set(i, float(val))
                continue

            rows = len(value)

            for i, values in enumerate(value):
                for j, val in enumerate(values):
                    weight_value.set(i * rows + j, float(val))

    def state_dict(self) -> Dict:
        raise NotImplementedError

    def load_state_dict(self, state_dict: Dict):
        raise NotImplementedError

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return draw_model(self, filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs)


class AbstractEvaluator:
    def __init__(self, backend: Backend, template: Template, settings: Settings):
        self.settings = settings.create_proxy()
        self.backend = backend
        self.dataset: Optional[BuiltDataset] = None

        self.neuralogic_model = template.build(backend, settings)

        if backend != Backend.PYG:
            self.neuralogic_model.set_hooks(template.hooks)

    def set_dataset(self, dataset: Union[Dataset, BuiltDataset]):
        self.dataset = self.build_dataset(dataset)

    def build_dataset(self, dataset: Union[Dataset, BuiltDataset], file_mode: bool = False):
        if isinstance(dataset, Dataset):
            return self.neuralogic_model.build_dataset(dataset, file_mode)
        return dataset

    @property
    def model(self) -> AbstractNeuraLogic:
        return self.neuralogic_model

    def train(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def state_dict(self) -> Dict:
        pass

    def load_state_dict(self, state_dict: Dict):
        pass

    def reset_parameters(self):
        self.neuralogic_model.reset_parameters()

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return self.neuralogic_model.draw(
            filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs
        )



from typing import List

from neuralogic.core import Relation, Template, Var, Aggregation, Activation, Metadata
from neuralogic.utils.templates.modules import AbstractModule


class SAGEConv(AbstractModule):
    def build(
        self, template: Template, layer_count: int, previous_names: List[str], feature_name: str, edge_name: str
    ) -> str:
        name = f"l{layer_count}_gsage" if self.name is None else self.name
        previous_name = feature_name if len(previous_names) == 0 else previous_names[-1]

        head_atom = Relation.get(name)(Var.X)[self.out_channels, self.in_channels]

        layer = head_atom <= (Relation.get(previous_name)(Var.Y), Relation.get(edge_name)(Var.Y, Var.X))
        template.add_rule(layer | Metadata(aggregation=Aggregation.AVG, activation=Activation.IDENTITY))

        layer = Relation.get(name)(Var.X)[self.out_channels, self.in_channels] <= Relation.get(previous_name)(Var.X)
        template.add_rule(layer | Metadata(activation=Activation.IDENTITY))

        template.add_rule(Relation.get(name) / 1 | Metadata(activation=self.activation))

        return name



from typing import Optional, List, Union, Tuple, Sequence, Iterable

import numpy as np

from neuralogic.core.constructs.factories import Relation
from neuralogic.core.constructs.atom import BaseAtom, WeightedAtom
from neuralogic.core.constructs.rule import Rule


DatasetEntries = Union[BaseAtom, WeightedAtom, Rule]


class Data:
    """
    Stores a learning example in the form of a tensor numeric representation instead of a rule based representation.
    """

    def __init__(
        self,
        x: Sequence = None,
        edge_index: Sequence = None,
        edge_attr: Optional[Sequence] = None,
        y_mask: Sequence = None,
        y: Union[Sequence, float, int] = None,
    ):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.y = y
        self.y_mask = y_mask

    @staticmethod
    def get_query(
        y, output_name: str = "predict", one_hot_encode_labels: bool = False, max_classes=1, index: Optional[int] = None
    ):
        relation = Relation.get(output_name)

        if index is not None:
            relation = relation(index)

        if one_hot_encode_labels:
            vector = y

            if not isinstance(y, (Sequence, np.ndarray)) or len(y) != max_classes:
                vector = np.zeros((max_classes,))
                vector[y[0] if isinstance(y, (Sequence, np.ndarray)) else y] = 1
            query = relation[vector]
        elif not isinstance(y, Iterable) or (getattr(y, "shape", -1) in (1, tuple())):
            query = relation[float(y)]
        elif len(y) == 1 and not isinstance(y[0], Sequence):
            query = relation[float(y[0])]
        else:
            if isinstance(y, (Sequence, np.ndarray)):
                query = relation[y]
            else:
                query = relation[y.detach().numpy()]
        return query

    def to_logic_form(
        self,
        feature_name: str = "node_feature",
        edge_name: str = "edge",
        output_name: str = "predict",
        one_hot_encode_labels=False,
        one_hot_decode_features=False,
        max_classes=1,
    ) -> Tuple:
        if self.y_mask is not None:
            query = []

            for i in self.y_mask:
                query.append(Data.get_query(self.y[int(i)], output_name, one_hot_encode_labels, max_classes, int(i)))
        else:
            query = Data.get_query(self.y, output_name, one_hot_encode_labels, max_classes)

        if self.edge_attr is None:
            example = [
                Relation.get(edge_name)(int(u), int(v))[1].fixed()
                for u, v in zip(self.edge_index[0], self.edge_index[1])
            ]
        elif isinstance(self.edge_attr, np.ndarray):
            example = [
                Relation.get(edge_name)(int(u), int(v))[w if w.size == 1 else w].fixed()
                for u, v, w in zip(self.edge_index[0], self.edge_index[1], self.edge_attr)
            ]
        elif isinstance(self.edge_attr, (Sequence, np.ndarray)):
            example = [
                Relation.get(edge_name)(int(u), int(v))[
                    w if len(w) == 1 and isinstance(w[0], float, int) else w
                ].fixed()
                for u, v, w in zip(self.edge_index[0], self.edge_index[1], self.edge_attr)
            ]
        else:
            example = [
                Relation.get(edge_name)(int(u), int(v))[w if w.size == 1 else w].fixed()
                for u, v, w in zip(self.edge_index[0], self.edge_index[1], self.edge_attr.detach().numpy())
            ]

        if one_hot_decode_features:
            if isinstance(self.x, (list, np.ndarray)):
                for i, features in enumerate(self.x):
                    class_ = np.argmax(features)
                    example.append(Relation.get(f"{feature_name}_{class_}")(i)[1].fixed())
            else:
                for i, features in enumerate(self.x):
                    class_ = np.argmax(features)
                    example.append(Relation.get(f"{feature_name}_{class_}")(i)[1].fixed())
        else:
            if isinstance(self.x, np.ndarray):
                for i, features in enumerate(self.x):
                    example.append(
                        Relation.get(feature_name)(i)[features[0] if features.size == 1 else features].fixed()
                    )
            elif isinstance(self.x, list):
                for i, features in enumerate(self.x):
                    example.append(
                        Relation.get(feature_name)(i)[
                            features[0] if len(features) == 1 and isinstance(features[0], (int, float)) else features
                        ].fixed()
                    )
            else:
                for i, features in enumerate(self.x):
                    weight = features.detach().numpy()
                    example.append(Relation.get(feature_name)(i)[weight[0] if weight.size == 1 else weight].fixed())
        return query, example

    @staticmethod
    def from_pyg(data) -> List["Data"]:
        """
        Converts a PyTorch Geometric Data instance into a list of PyNeuraLogic :py:class:`~Data` instances.
        The conversion supports :code:`train_mask`, :code:`test_mask` and :code:`val_mask` attributes -
        for each mask the conversion yields a new data instance.

        :param data: The PyTorch Geometric Data instance
        :return: The list of PyNeuraLogic Data instances
        """
        data_list = []

        if hasattr(data, "train_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.train_mask, data.y))
        if hasattr(data, "test_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.test_mask, data.y))
        if hasattr(data, "val_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.val_mask, data.y))
        if len(data_list) == 0:
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, None, data.y))

        return data_list


class Dataset:
    def __init__(
        self,
        *,
        data: Optional[List[Data]] = None,
        examples_file: Optional[str] = None,
        queries_file: Optional[str] = None,
        one_hot_encode_labels: bool = False,
        one_hot_decode_features: bool = False,
        number_of_classes: int = 1,
        feature_name: str = "node_feature",
        edge_name: str = "edge",
        output_name: str = "predict",
    ):
        self.file_sources = False

        if examples_file is not None or queries_file is not None:
            if data is not None:
                raise Exception
            self.file_sources = True

        self.data = data
        self.examples_file = examples_file
        self.queries_file = queries_file

        self.examples: List[DatasetEntries] = []
        self.queries: List[DatasetEntries] = []

        self.one_hot_decode_features = one_hot_decode_features
        self.one_hot_encode_labels = one_hot_encode_labels
        self.number_of_classes = number_of_classes

        self.feature_name: str = feature_name
        self.edge_name: str = edge_name
        self.output_name: str = output_name

    def __len__(self):
        if self.data is not None:
            return len(self.data)
        if self.file_sources:
            return -1
        return max(len(self.examples), len(self.queries))

    def __getitem__(self, item):
        pass

    def add_example(self, example):
        self.add_examples([example])

    def add_examples(self, examples: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.examples.extend(examples)

    def set_examples(self, examples: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.examples = examples

    def add_query(self, query):
        self.add_queries([query])

    def add_queries(self, queries: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.queries.extend(queries)

    def set_queries(self, queries: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.queries = queries

    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        for data in self.data:
            query, examples = data.to_logic_form(
                self.feature_name,
                self.edge_name,
                self.output_name,
                self.one_hot_encode_labels,
                self.one_hot_decode_features,
                self.number_of_classes,
            )

            queries_fp.write(f"{query}{sep}")
            examples_fp.write(f"{','.join(example.to_str(False) for example in examples)}.{sep}")

    def dump_to_file(
        self,
        queries_filename: str,
        examples_filename: str,
        sep: str = "\n",
    ):
        with open(queries_filename, "w") as queries_fp:
            with open(examples_filename, "w") as examples_fp:
                self.dump(queries_fp, examples_fp, sep)



from typing import List

from neuralogic.core import Template
from neuralogic.core.enums import Aggregation, Activation


class AbstractModule:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.SUM,
        name=None,
        has_edge_attrs=True,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.has_edge_attrs = has_edge_attrs
        self.aggregation = aggregation
        self.activation = activation
        self.name = name

    def build(
        self, template: Template, layer_count: int, previous_names: List[str], feature_name: str, edge_name: str
    ) -> str:
        pass



from typing import List

from neuralogic.core import Template
from neuralogic.utils.templates.modules import AbstractModule

from neuralogic.utils.templates.modules.gcn import GCNConv
from neuralogic.utils.templates.modules.gsage import SAGEConv
from neuralogic.utils.templates.modules.gin import GINConv
from neuralogic.utils.templates.modules.pooling import GlobalPooling


class TemplateList:
    """TemplateList is a collection of pre-defined modules"""

    def __init__(
        self,
        modules: List[AbstractModule],
        edge_name: str = "_edge",
        feature_name: str = "node_feature",
        output_name: str = "predict",
    ):
        self.modules = modules
        self.edge_name = edge_name
        self.feature_name = feature_name
        self.output_name = output_name

    def build(self, template: Template):
        if len(self.modules) == 0:
            return

        previous_names = []
        output_layer = len(self.modules)

        for i, component in enumerate(self.modules):
            name = component.name

            if i == output_layer - 1:
                component.name = self.output_name

            previous_names.append(component.build(template, i + 1, previous_names, self.feature_name, self.edge_name))

            component.name = name



from typing import Optional, Union

import torch.nn.functional as F
import torch

from neuralogic.core.error_function import ErrorFunctionNames
from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset, Backend, Settings, Optimizer, Dataset


class PyGEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: torch.optim.SGD(param, lr=rate),
        Optimizer.ADAM: lambda param, rate: torch.optim.Adam(param, lr=rate),
    }

    error_functions = {
        str(ErrorFunctionNames.MSE): F.mse_loss,
        str(ErrorFunctionNames.CROSSENTROPY): F.cross_entropy,
    }

    def __init__(self, template: Template, settings: Settings):
        super().__init__(Backend.PYG, template, settings)

    def train(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        # dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = str(self.settings.error_function)
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in PyGEvaluator.trainers:
            raise NotImplementedError
        if error_function not in PyGEvaluator.error_functions:
            raise NotImplementedError

        trainer = PyGEvaluator.trainers[optimizer](
            self.neuralogic_model.module_list.parameters(),
            self.settings.learning_rate,
        )
        error_function = PyGEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0

                for data in dataset.data:
                    self.neuralogic_model.train()
                    trainer.zero_grad()

                    out = self.neuralogic_model(x=data.x, edge_index=data.edge_index)

                    if data.y_mask is None:
                        loss = error_function(out, data.y)
                    else:
                        loss = error_function(out[data.y_mask], data.y[data.y_mask])
                    loss.backward()
                    trainer.step()

                    seen_instances += 1
                    total_loss += float(loss)
                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        self.neuralogic_model.train(mode=False)

        # dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for data in dataset.data:
                self.neuralogic_model.train(mode=False)
                out = self.neuralogic_model(x=data.x, edge_index=data.edge_index)
                results = (out[data.y_mask], data.y[data.y_mask])

                yield results

        if generator:
            return _test()
        return list(_test())



from neuralogic.core.constructs.factories import Var, Term, Relation, V, T, R
from neuralogic.core.template import Template
from neuralogic.core.builder import BuiltDataset
from neuralogic.core.dataset import Dataset, Data
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.settings import Settings, SettingsProxy
from neuralogic.core.enums import Activation, ActivationAgg, Aggregation, Optimizer, Initializer, Backend
